---
--- monadic interpreters a la Wadler
--- see Phil Wadler's paper "The essence of functional programming" POPL 92
---

import Control.Monad

---
--- Nondeterministic choice monad
---
data AltM a = Alt [a]

runAltM :: AltM a -> [a]
runAltM (Alt l) = l

instance Monad AltM where
  return x = Alt [x]
  m >>= k  = Alt [b | a <- runAltM m, b <- runAltM (k a) ]
  
instance MonadPlus AltM where
  mzero = Alt []
  mplus (Alt xs) (Alt ys) = Alt (xs ++ ys)

instance Show a => Show (AltM a) where
  show (Alt m) = show m

---
--- data types for interpreter
---
type Name = String

data Term = Var Name
          | Lit Int
          | Lam Name Term 
          | App Term Term 
          | Add Term Term 
          | Fail            -- added
          | Or Term Term    -- added
            
data Value = Wrong 
           | Num Int
           | Fun (Value -> AltM Value)
             
type Environment = [(Name, Value)]

instance Show Value where
  show Wrong = "<wrong>"
  show (Num i) = show i
  show (Fun _) = "<function>"


--- interpreter

eval  :: Term -> Environment -> AltM Value
eval (Var v)   e = lookup_ v e
eval (Lit x)   e = return (Num x)
eval (Lam x t) e = return (Fun (\a -> eval t ((x, a):e)))
eval (App t u) e = do f <- eval t e
                      a <- eval u e
                      apply f a
eval (Add t u) e = do a <- eval t e
                      b <- eval u e
                      add a b
eval Fail      e = mzero                        -- added
eval (Or t u)  e = eval t e `mplus` eval u e    -- added

run :: Term -> AltM Value
run t = eval t []


--- auxiliary functions

lookup_  :: Name -> Environment -> AltM Value
lookup_ x []          = return Wrong 
lookup_ x ((y, v):es) = if x == y then return v else lookup_ x es

apply :: Value -> Value -> AltM Value
apply (Fun k) a = k a
apply _       _ = return Wrong

add :: Value -> Value -> AltM Value
add (Num x) (Num y) = return (Num (x + y))
add _       _       = return Wrong


-- terms for testing

test = [t0, t1, t2]

-- (\x. x + x) (10 + 11)
t0 = (App (Lam "x" (Add (Var "x") (Var "x"))) (Add (Lit 10) (Lit 11)))

-- (\x. x + x) (10 | 11)
t1 = (App (Lam "x" (Add (Var "x") (Var "x"))) (Or (Lit 10) (Lit 11)))

-- (7 | 10) + Fail
t2 = Add (Or (Lit 7) (Lit 10)) Fail

-- (\x y. x + x + y) (10 | 11) ()
t3l = Lam "x" (Lam "y" (Add (Add (Var "x") (Var "x")) (Var "y")))
t3 = App (App t3l (Or (Lit 10) (Lit 20))) (Or (Lit 1) (Lit 5))


